import math
import quadtree

from .Bbox import Bbox

class TileBounds(object):
    maxzoom = 21
    worldwidth = 360.0
    worldheight = 180.0
    min_tile_width = worldwidth / pow(2, maxzoom)
    min_tile_height = worldheight / pow(2, maxzoom)

    def __init__(self, gridcode=''):
        """
        gridcode value of '' (empty string) is the world tile Bbox(-180, -90, 180, 90)
        """
        self.gridcode = gridcode

    @staticmethod
    def world_bounds():
        return TileBounds()

    @classmethod
    def from_string(cls, s):
        return cls(gridcode=s)

    @classmethod
    def from_bbox_best_fit(cls, bbox):
        """
        Return a TileBounds that is about the same size as the bbox and contains the center point
        """
        # get the zoom level based on the width of the given bounding box
        level = min(cls.maxzoom,
                    max(0,
                        int(math.floor(min(
                            math.log(cls.worldwidth / max(bbox.width, cls.min_tile_width), 2),
                            math.log(cls.worldheight / max(bbox.height, cls.min_tile_height), 2))))
                    )
        )
        return cls(gridcode=quadtree.encode(bbox.latmin + bbox.height / 2, bbox.lonmin + bbox.width / 2, level))

    @classmethod
    def from_point(cls, lon, lat, zoom_level):
        assert 0 <= zoom_level <= cls.maxzoom
        # lon==180 is not in the quad tree, because tiles only contian points on the left and bottom edges
        # points on the top and right edges belon to the adjacent tile
        # quadtree will throw if you give it lon==180, so we map it to -180 here
        if lon == 180.0:
            lon = -180.0

        # lat==90 is also a problem because this is always on the top egde of the tile, but we can't wrap
        # around, so instead we fudge it down by half the height of the smallest possible tile
        if lat == 90.0:
            lat = 90 - (cls.min_tile_height / 2)
        return cls(gridcode=quadtree.encode(lat=lat, lon=lon, precision=zoom_level))

    def get_bbox(self):
        bounds = quadtree.bbox(self.gridcode)
        return Bbox(bounds['w'], bounds['s'], bounds['e'], bounds['n'])

    def get_center(self):
        """
        Get the point at the center of the tile
        """
        return self.get_bbox().center

    def get_parent(self):
        if self.gridcode:
            return TileBounds(gridcode=self.gridcode[:-1])

    def get_children(self):
        return [TileBounds(gridcode=self.gridcode + digit) for digit in '0123']

    def get_ancestors(self):
        return [TileBounds(gridcode=self.gridcode[:-i]) for i in range(1, len(self.gridcode) + 1)]

    def contains_point(self, lon, lat):
        return self == TileBounds.from_point(lon, lat, self.zoom_level)

    @property
    def zoom_level(self):
        return len(self.gridcode)

    def __str__(self):
        return self.gridcode

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, str(self))

    def __eq__(self, other):
        return str(self) == str(other)



class Bbox(object):
    def __init__(self, lonmin=-180, latmin=-90, lonmax=180, latmax=90):
        self.latmin = float(min(latmin, latmax))
        self.latmax = float(max(latmin, latmax))
        self.lonmin = float(min(lonmin, lonmax))
        self.lonmax = float(max(lonmin, lonmax))

    @classmethod
    def fromstring(cls, s):
        """Initialize Bbox from a string formatted by str()
        """
        return cls(*[float(item) for item in s.split(",")])

    def __str__(self):
        def f(v):
            # Format v as javascript toString():
            if (v == 0.0): return "0"
            return str(v).rstrip("0").rstrip(".")

        # left,bottom,right,top - all according to openlayers :)
        return "%s,%s,%s,%s" % (f(self.lonmin), f(self.latmin), f(self.lonmax), f(self.latmax))

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self))

    @property
    def height(self):
        return self.latmax - self.latmin

    @property
    def width(self):
        return self.lonmax - self.lonmin

    def contains(self, lon, lat):
        """
        Return True if the given point is inside the bbox, False if not.
        Note that a point exactly on the left or bottom edge of the bbox is consider inside, while a point
        exactly on the right or top edge is not.  the point on the lower left corner is inside (True), all 3 other
        corner points are outside (False)
        """
        return self.lonmin <= lon < self.lonmax and self.latmin <= lat < self.latmax

    def __eq__(self, other):
        return self.lonmin == other.lonmin and \
                self.lonmax == other.lonmax and \
                self.latmin == other.latmin and \
                self.latmax == other.latmax

    def asdict(self):
        return self.__dict__

    @property
    def center (self):
        """
        Get the point at center of the bbox

        :rtype : dict(lon=lon, lat=lat)
        """
        return dict(lon=self.lonmin + (self.width/2), lat=self.latmin+ (self.height/2))



import unittest2
from .. import Bbox


class TileTest(unittest2.TestCase):
    def test_bbox(self):
        b = Bbox(1.2345678, -2.34, 3, 4)
        self.assertEqual('Bbox(1.2345678,-2.34,3,4)', repr(b))
        self.assertEqual('1.2345678,-2.34,3,4', str(b))
        self.assertEqual('Bbox(1.2345678,-2.34,3,4)', repr(eval(repr(b))))

        self.assertDictEqual(Bbox(2,2,4,4).center, dict(lat=3.0, lon=3.0))



from datetime import datetime
from vectortile import TypedMatrix


class Tile(object):
    def __init__(self, typedmatrix_str, meta=None):
        """

        :param typedmatrix_str: a packed string as returned by TypedMatrix.pack()
        """
        self.typedmatrix_str = typedmatrix_str
        if not meta:
            meta = dict()
        self.meta = meta

    @staticmethod
    def fromdata(data, meta=None, columns=None):
        """

        :param data: a dict or list of dicts suitable for TypedMatrix.pack()
        :param meta: a dict of meta data fields
        :return: a Tile object
        """
        if not meta:
            meta = dict()
        return Tile(TypedMatrix.pack(data, extra_header_fields=meta, columns=columns, orientation='columnwise'), meta=meta)

    @staticmethod
    def nodata(complete_ancestor_bounds=None):
        meta = dict(nodata=True)
        if complete_ancestor_bounds:
            meta['complete_ancestor'] = str(complete_ancestor_bounds.get_bbox())
        return Tile.fromdata(data=dict(), meta=meta)

    def is_nodata(self):
        return 'nodata' in self.meta

    def complete_ancestor(self):
        return self.meta.get('complete_ancestor')

    @staticmethod
    def timestamp(dt=datetime.now()):
        """
        Create a coded UTC timestamp from a datetime using the internal coding format used in packed tiles
        """
        return TypedMatrix.get_utc_timestamp(dt)

    def unpack(self):
        return TypedMatrix.unpack(self.typedmatrix_str)

    def __str__(self):
        return "%s" % self.typedmatrix_str

    def __repr__(self):
        return "%s(%s,meta=%s)" % (self.__class__.__name__, repr(self.typedmatrix_str), repr(self.meta))

    @property
    def size(self):
        return len(self.typedmatrix_str)

    def asdict(self):
        header, data = self.unpack()
        header['data'] = data
        return header



import unittest2
from vectortile import TileBounds, Bbox


class TileTest(unittest2.TestCase):
    def test_tile_bounds(self):
        # default tile bounds is the world tile
        self.assertEqual(TileBounds().get_bbox(), Bbox())

        # make sure Bounds to Bbox and back is reflexive
        bounds = TileBounds('12321')
        bbox = bounds.get_bbox()
        self.assertEqual(bbox, Bbox(56.25, -11.25, 67.5, -5.625))
        self.assertEqual(TileBounds.from_bbox_best_fit(bbox), bounds)

        # test __str__ and __repr__
        self.assertEqual("TileBounds('12321')", repr(eval(repr(bounds))))

        # parents and children
        self.assertEqual(bounds.get_parent(), '1232')
        self.assertListEqual(bounds.get_children(), ['123210', '123211', '123212', '123213'])
        self.assertListEqual(bounds.get_ancestors(), ['1232', '123', '12', '1', ''])

        # from point
        self.assertEqual(TileBounds.from_point(lon=10, lat=10, zoom_level=0), TileBounds())
        self.assertEqual(TileBounds.from_point(lon=10, lat=10, zoom_level=1).get_bbox(), Bbox(0, 0, 180, 90))

        # from bbox
        self.assertEqual(TileBounds.from_bbox_best_fit(Bbox(0, 0, 180, 90)).get_bbox(), Bbox(0, 0, 180, 90))
        self.assertEqual(TileBounds.from_bbox_best_fit(Bbox(56.25, -11.25, 67.5, -5.625)).get_bbox(),
                         Bbox(56.25, -11.25, 67.5, -5.625))
        self.assertEqual(TileBounds.from_bbox_best_fit(Bbox(56, -11.25, 68, -5.625)).get_bbox(),
                         Bbox(+045.00000, -011.25000, +067.50000, +000.00000))

        # center
        self.assertDictEqual(TileBounds().get_center(), dict(lat=0.0, lon=0.0))
        self.assertDictEqual(TileBounds('0').get_center(), dict(lat=-45.0, lon=-90.0))

        # boundary cases
        self.assertListEqual(TileBounds().get_ancestors(), [])
        self.assertEqual(TileBounds.from_point(lon=180, lat=0, zoom_level=1).get_bbox(),
                         Bbox(-180, 0, 0, 90))
        self.assertEqual(TileBounds.from_point(lon=0, lat=90, zoom_level=1).get_bbox(),
                         Bbox(0, 0, 180, 90))




import unittest2
from .. import TypedMatrix
from datetime import datetime, timedelta


class TypedMatrixTest(unittest2.TestCase):
    def test_get_columns(self):
        data = {'A': 1}
        cols = TypedMatrix.get_columns(data)
        self.assertEqual(1, len(cols))
        self.assertEqual('A', cols[0]['name'])
        self.assertEqual('Float32', cols[0]['type'])

    def test_pack(self):
        data = {'A': 1}
        packed_str = TypedMatrix.pack(data)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual(1, header['length'])
        self.assertEqual(1, data[0]['A'])
        data = {'A': 1, 'B': 2}
        packed_str = TypedMatrix.pack(data)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual(1, header['length'])
        self.assertEqual(1, data[0]['A'])
        self.assertEqual(2, data[0]['B'])
        data = [{'A': 1, 'B': 2}, {'A': 11, 'B': 22}]
        packed_str = TypedMatrix.pack(data)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual(2, header['length'])
        self.assertEqual(1, data[0]['A'])
        self.assertEqual(22, data[1]['B'])

    def test_header(self):
        data = {'A': 1, 'B': 2}
        header = {'H': 'extra'}
        packed_str = TypedMatrix.pack(data, header)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual('extra', header['H'])
        packed_str = TypedMatrix.pack(data, extra_header_fields=header, orientation='columnwise')
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual('extra', header['H'])

    def test_get_packed_float_value(self):
        self.assertEqual(TypedMatrix.get_packed_float_value(1.0), 1.0)
        self.assertEqual(TypedMatrix.get_packed_float_value(1234567890.0), 1234567936.0)

    def test_get_gmt_timestamp(self):
        self.assertEqual(TypedMatrix.get_utc_timestamp(datetime(2012, 1, 1)), 1325375946752.0)

    def test_datetime(self):
        data_in = dict(I=1, F=2.0, D=datetime(2014, 1, 1))
        data_out = data_in
        data_out['D'] = TypedMatrix.get_utc_timestamp(data_out['D'])

        packed_str = TypedMatrix.pack(data_in)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertDictEqual(dict(name='D', type='Float32'), header['cols'][0])
        self.assertDictEqual(dict(name='F', type='Float32'), header['cols'][1])
        self.assertDictEqual(dict(name='I', type='Float32'), header['cols'][2])
        self.assertDictEqual(data[0], data_out)

    def test_unsupported_type(self):
        data_in = dict(I=1, F=2.0, D=datetime(2014, 1, 1), unsupported=timedelta(1))
        self.assertRaises(TypeError, TypedMatrix.pack, data_in)

    def test_bad_format(self):
        self.assertRaises(AssertionError, TypedMatrix.unpack, packed_str=b'not valid')

    def test_columnwise(self):
        data_in = [{'A': 1, 'B': 2}, {'A': 11, 'B': 22}, {'A': 111, 'B': 222}]
        packed_str = TypedMatrix.pack(data_in, orientation='columnwise')
        header, data_out = TypedMatrix.unpack(packed_str)
        self.assertEqual(3, header['length'])
        self.assertListEqual(data_in, data_out)

        data_in = [{'AA': 1, 'BB': 2}, {'AA': 11, 'BB': 22}]
        packed_str = TypedMatrix.pack(data_in, orientation='columnwise')
        header, data_out = TypedMatrix.unpack(packed_str)
        self.assertEqual(2, header['length'])
        self.assertListEqual(data_in, data_out)

    def test_empty(self):
        data = []
        packed_str = TypedMatrix.pack(data)
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual(0, header['length'])
        self.assertEqual(len(data), 0)

        packed_str = TypedMatrix.pack(data, orientation='columnwise')
        header, data = TypedMatrix.unpack(packed_str)
        self.assertEqual(0, header['length'])
        self.assertEqual(len(data), 0)



"""
TypedMatrix.py - read/write TypedMatrix format
"""

magic = b'tmtx' # Must be four bytes long!
version = 2

import io
import struct
import json
import calendar
from datetime import datetime
try:
    import pandas
except:
    pandas = None

# Header Structure
# {
#   length: count of rows of data
#   version: format version
#   cols: array of column definitions
#       [{}
#       ]
# }

typemap = {
    int: 'Float32',
    float: 'Float32',
    datetime: 'Float32',
}
if pandas:
    typemap[pandas.Timestamp] = 'Float32'
typeformatmap = {
    'Float32': 'f',
}
typedefaultmap = {
    'Float32': 0.0,
}

def get_columns(data):
    """
    Gets the column definitions implicit in a dict or list of dicts.
    If any field which has a datatype that is not in typemap, thows TypeException
    """
    # make data iterable
    if type(data) is dict:
        data = [data]

    cols = {}
    for i, d in enumerate(data):
        for key, value in d.items():
            t = type(value)
            if t not in typemap:
                raise TypeError ('TypedMatrix: "%s" is not a supported type in field "%s"' % (type(value), key))
            if key not in cols:
                cols[key] = {'name': key, 'type': typemap[t]}
    cols = list(cols.values())
    cols.sort(key=lambda a: a['name'])
    return cols


def _datetime2timestamp(dt):
    return calendar.timegm(dt.utctimetuple()) * 1000.0


def conv(data, t, default):
    if 'Float32' == t:
        if type(data) is datetime:
            fn = _datetime2timestamp
        else:
            fn = float
    else:
        assert False, 'Unknown conversion type %s' % t

    # noinspection PyBroadException
    try:
        return fn(data)
    except:
        return default


def row_fmt(columns):
    return '<' + ''.join(typeformatmap[col['type']] for col in columns)


def pack(data, extra_header_fields=None, columns=None, orientation='rowwise'):
    """
    Pack a dict or list of dicts into a TypedMatrix binary packed string
    If a list of columns is not given, the columns are derived from the data using get_columns()
    extra_header_fields can supply an optional dict with additional fields to be included in the packed header

    orientation can be 'rowwise' or 'columnwise'.   For row-wise orientation, the typedmatrix will stored as a list of rows.
    for column-wise oreintation it is stored as a list of columns.
    """

    if hasattr(data, "to_dict"):
        data = data.to_dict('records')
    
    # make data iterable
    if type(data) is dict:
        data = [data]

    if extra_header_fields:
        header = dict(extra_header_fields)
    else:
        header = dict()

    header['length'] = len(data)
    if not columns:
        columns = get_columns(data)
    header['cols'] = columns

    if 'colsByName' in header:
        colsByName = header.pop("colsByName")
        for col in header['cols']:
            if col['name'] in colsByName:
                col.update(colsByName[col['name']])

    header['version'] = version
    if orientation is None:
        raise ValueError ('TypedMatrix: unknown orientation %s' % orientation)
    header['orientation'] = orientation

    f = io.BytesIO()
    headerstr = json.dumps(header).encode("utf-8")
    paddinglen = (4 - len(headerstr) % 4) % 4
    headerstr += (b" " * paddinglen)


    # write "magic" file format token at the start
    f.write(struct.pack('<%ss' % len(magic), magic))
    f.write(struct.pack("<i", len(headerstr)))
    f.write(headerstr)

    colspecs = [{'name': col['name'], 'type': col['type'], 'default': typedefaultmap[col['type']]} for col in columns]

    if orientation == 'rowwise':
        for d in data:
            f.write(struct.pack(
                row_fmt(columns),
                *[conv(d[colspec['name']], colspec['type'], colspec['default'])
                  for colspec in colspecs]))
    else:
        for colspec in colspecs:
            f.write(struct.pack(
                '<%s%s' % (len(data),typeformatmap[colspec['type']]),
                *[conv(d[colspec['name']], colspec['type'], colspec['default'])
                  for d in data]))

    return f.getvalue()


def _struct_read (f, t, n=1):
    fmt = '<%s%s' % (n, t)
    result = struct.unpack(fmt, f.read(struct.calcsize(fmt)))
    if n == 1:
        return result[0]
    return result


def unpack(packed_str):
    """
    Unpack a binary packed string containing a TypedMatrix.   Returns a tuple of header and data
    header is a dict and data is a list of dicts
    """
    f = io.BytesIO(packed_str)

    # read "magic" file format token
    token = b''.join(_struct_read(f, 'c', len(magic)))
    assert(token == magic)

    header_len = _struct_read(f,'i')
    header = json.loads(f.read(header_len).decode("utf-8"))

    assert (header['version'] == version)  # only supports one version right now

    if header['orientation'] == 'rowwise':
        fmt = row_fmt(header['cols'])
        data = []
        col_names = [col['name'] for col in header['cols']]
        for i in range(0, header['length']):
            data.append(dict(zip(col_names, struct.unpack(fmt, f.read(struct.calcsize(fmt))))))
    elif header['orientation'] == 'columnwise':
        col_data = []
        col_names = [col['name'] for col in header['cols']]
        length = header['length']
        for col in header['cols']:
            col_data.append(_struct_read(f, typeformatmap[col['type']], length))
        col_indexes = range(0, len(col_names))
        if length > 1:
            data = [dict(zip(col_names, [col_data[c][i] for c in col_indexes])) for i in range(0, length)]
        elif length == 1:
            data = [dict(zip(col_names, [col_data[c] for c in col_indexes]))]
        else:
            data = []
    else:
        raise Exception("Unsupported orientation")

    return header, data


def get_packed_float_value(f):
    """
    Get a float value that has been packed using Float32.   This is useful to get the actual precision used
    in the packed array for a float column, since Float32 precision is limited to about 9 digits and
    there is no native Float32 in python.
    """
    return struct.unpack('<f', struct.pack("<f", float(f)))[0]


def get_utc_timestamp(dt=datetime.now()):
    """
    Create a timestamp in milliseconds as a Float32 from a datetime
    """
    return get_packed_float_value(_datetime2timestamp(dt))



import unittest2
from .. import Tile


class TileTest(unittest2.TestCase):
    def test_tile(self):
        d = {'A': 1}
        h = {'source': 'test_Tile'}
        tile = Tile.fromdata(d, meta=h)
        header, data = tile.unpack()
        self.assertEqual(header['length'], 1)
        self.assertEqual(header['source'], h['source'])
        self.assertEqual(data[0]['A'], d['A'])
        self.assertEqual(repr(tile), repr(eval(repr(tile))))

    def test_empty(self):
        d = []
        h = {'source': 'test_Tile'}
        tile = Tile.fromdata(d, meta=h)
        header, data = tile.unpack()
        self.assertEqual(header['length'], 0)
        self.assertEqual(len(data), 0)




#!/usr/bin/env python

import os.path
import distutils.core

# Don't import vectortile here, that would create loops. Just load
# this one file that doesn't import anything.

versionpy = os.path.join(os.path.split(os.path.abspath(__file__))[0], "vectortile/version.py")
version = {}
with open(versionpy) as f:
    exec(f.read(), version)

with open('README.md') as f:
    readme = f.read()

with open('requirements.txt') as f:
    requirements = [l for l in f.readlines() if l]

distutils.core.setup(
    name='vectortile',
    description="A set of classes for managing tiles of geospatial vector data",
    long_description_content_type="text/markdown",
    long_description=readme,
    packages=[
        'vectortile',
        'vectortile.tests'
    ],
    install_requires=requirements,
    version=version['__version__'],
    author=version['__author__'],
    author_email=version['__author_email__'],
    url=version['__source__'],
    license=version['__license__']
)
